@@ -110,7 +110,7 @@ libxfs_trans_roll(
 	/*
 	 * Commit the current transaction.
 	 * If this commit failed, then it'd just unlock those items that
-	 * are not marked ihold. That also means that a filesystem shutdown
+	 * are marked to be released. That also means that a filesystem shutdown
 	 * is in progress. The caller takes the responsibility to cancel
 	 * the duplicate transaction that gets returned.
 	 */
@@ -247,27 +247,6 @@ libxfs_trans_iget(
 	return 0;
 }
 
-void
-libxfs_trans_iput(
-	xfs_trans_t		*tp,
-	xfs_inode_t		*ip,
-	uint			lock_flags)
-{
-	xfs_inode_log_item_t	*iip;
-
-	if (tp == NULL) {
-		libxfs_iput(ip, lock_flags);
-		return;
-	}
-
-	ASSERT(ip->i_transp == tp);
-	iip = ip->i_itemp;
-	ASSERT(iip != NULL);
-	xfs_trans_del_item(&iip->ili_item);
-
-	libxfs_iput(ip, lock_flags);
-}
-
 void
 libxfs_trans_ijoin(
 	xfs_trans_t		*tp,
@@ -301,28 +280,12 @@ libxfs_trans_ijoin_ref(
 	ASSERT(ip->i_itemp != NULL);
 
 	xfs_trans_ijoin(tp, ip, lock_flags);
-	ip->i_itemp->ili_lock_flags = lock_flags;
 
 #ifdef XACT_DEBUG
 	fprintf(stderr, "ijoin_ref'd inode %llu, transaction %p\n", ip->i_ino, tp);
 #endif
 }
 
-void
-libxfs_trans_ihold(
-	xfs_trans_t		*tp,
-	xfs_inode_t		*ip)
-{
-	ASSERT(ip->i_transp == tp);
-	ASSERT(ip->i_itemp != NULL);
-
-	ip->i_itemp->ili_lock_flags = 1;
-
-#ifdef XACT_DEBUG
-	fprintf(stderr, "ihold'd inode %llu, transaction %p\n", ip->i_ino, tp);
-#endif
-}
-
 void
 libxfs_trans_inode_alloc_buf(
 	xfs_trans_t		*tp,
@@ -702,7 +665,7 @@ inode_item_done(
 	if (!(iip->ili_fields & XFS_ILOG_ALL)) {
 		ip->i_transp = NULL;	/* disassociate from transaction */
 		iip->ili_flags = 0;	/* reset all flags */
-		goto ili_done;
+		return;
 	}
 
 	/*
@@ -712,7 +675,7 @@ inode_item_done(
 	if (error) {
 		fprintf(stderr, _("%s: warning - imap_to_bp failed (%d)\n"),
 			progname, error);
-		goto ili_done;
+		return;
 	}
 
 	XFS_BUF_SET_FSPRIVATE(bp, iip);
@@ -720,7 +683,7 @@ inode_item_done(
 	if (error) {
 		fprintf(stderr, _("%s: warning - iflush_int failed (%d)\n"),
 			progname, error);
-		goto ili_done;
+		return;
 	}
 
 	ip->i_transp = NULL;	/* disassociate from transaction */
@@ -728,16 +691,9 @@ inode_item_done(
 	XFS_BUF_SET_FSPRIVATE2(bp, NULL);	/* remove xact ptr */
 	libxfs_writebuf(bp, 0);
 #ifdef XACT_DEBUG
-	fprintf(stderr, "flushing dirty inode %llu, buffer %p (hold=%u)\n",
-			ip->i_ino, bp, iip->ili_lock_flags);
+	fprintf(stderr, "flushing dirty inode %llu, buffer %p\n",
+			ip->i_ino, bp);
 #endif
-ili_done:
-	if (iip->ili_lock_flags) {
-		iip->ili_lock_flags = 0;
-		return;
-	}
-	/* free the inode */
-	libxfs_iput(ip, 0);
 }
 
 static void
@@ -818,10 +774,6 @@ inode_item_unlock(
 	ip->i_transp = NULL;
 
 	iip->ili_flags = 0;
-	if (!iip->ili_lock_flags)
-		libxfs_iput(ip, 0);
-	else
-		iip->ili_lock_flags = 0;
 }
 
 /*
